library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;
use work.secded_32b_pkg.all;

--library smartfusion2;
--use smartfusion2.all;

entity sf2_test is
  generic (
    g_SIM_SEU_PERIOD : natural := 0
  );
  port (
    xtl   : in  std_logic;
    devrst_n : in std_logic;

    led1_o : out std_logic;
    led2_o : out std_logic;
    but_i : in std_logic;

    rx_i : std_logic;
    tx_o : out std_logic
  );
end sf2_test;

architecture behav of sf2_test is
  attribute syn_netlist_hierarchy : boolean;
  attribute syn_netlist_hierarchy of behav: architecture is false;
  
  signal FIC_0_AHB_S_HADDR     : std_logic_vector(31 downto 0);
  signal FIC_0_AHB_S_HMASTLOCK : std_logic;
  signal FIC_0_AHB_S_HREADY    : std_logic;
  signal FIC_0_AHB_S_HSEL      : std_logic;
  signal FIC_0_AHB_S_HSIZE     : std_logic_vector(1 downto 0);
  signal FIC_0_AHB_S_HTRANS    : std_logic_vector(1 downto 0);
  signal FIC_0_AHB_S_HWDATA    : std_logic_vector(31 downto 0);
  signal FIC_0_AHB_S_HWRITE    : std_logic;
  signal M3_RESET_N            : std_logic;
  signal POWER_ON_RESET_N      : std_logic;
  signal MMUART_0_RXD          : std_logic;
  signal FIC_0_AHB_S_HRDATA    : std_logic_vector(31 downto 0);
  signal FIC_0_AHB_S_HREADYOUT : std_logic;
  signal FIC_0_AHB_S_HRESP     : std_logic;
  signal MSS_RESET_N_M2F       : std_logic;
  signal LOCK                  : std_logic;
  signal MMUART_0_TXD          : std_logic;
  signal clk_100               : std_logic;

  signal rst_n : std_logic;

  type state_t is (INIT_DELAY,
                   READ_NVM,
                   WRITE_MEM,
                   DONE);
  signal state : state_t;
  signal counter : unsigned(27 downto 0);

  signal ahb_start, ahb_done, ahb_rw : std_logic;
  signal ahb_addr, ahb_wdata, ahb_rdata : std_logic_vector(31 downto 0);

  signal iahb_start : std_logic;
  signal iahb_addr : std_logic_vector(31 downto 0);

  type ahb_state_t is (STATE_AHB_IDLE, STATE_AHB_ADDR, STATE_AHB_DATA);
  signal ahb_state : ahb_state_t;

  --  IRAM log size in bytes.
  constant IRAM_LOG_SIZE : natural := 12;

  signal iram_addr : std_logic_vector(IRAM_LOG_SIZE - 1 downto 2);
  signal iram_we : std_logic;
  signal iram_data : std_logic_vector(31 downto 0);
  signal iram_ecc, iram_ecc_force : std_logic_vector(6 downto 0);

  signal cpu_rst_n : std_logic := '0';

  signal dwb_out : t_wishbone_master_out;
  signal dwb_in  : t_wishbone_master_in;

  signal cpu_err : std_logic;
begin
  inst_hydra: entity work.hydra_core
    generic map (
      g_iram_log_size => IRAM_LOG_SIZE,
      g_dram_log_size => 12,
      g_SIM_SEU_PERIOD => g_SIM_SEU_PERIOD
    )
    port map (
      clk_sys_i => clk_100,
      rst_n_i => rst_n,
      cpu_rst_n_i => cpu_rst_n,
      dwb_o => dwb_out,
      dwb_i => dwb_in,
      iram_addr_i => iram_addr,
      iram_data_i => iram_data,
      iram_we_i => iram_we,
      iram_ecc_i => iram_ecc,
      cpu_err_o => cpu_err
    );

  inst_uart: entity work.uart
    port map (
      FIC_0_AHB_S_HADDR     => FIC_0_AHB_S_HADDR,
      FIC_0_AHB_S_HMASTLOCK => FIC_0_AHB_S_HMASTLOCK,
      FIC_0_AHB_S_HREADY    => FIC_0_AHB_S_HREADY,
      FIC_0_AHB_S_HSEL      => FIC_0_AHB_S_HSEL,
      FIC_0_AHB_S_HSIZE     => FIC_0_AHB_S_HSIZE,
      FIC_0_AHB_S_HTRANS    => FIC_0_AHB_S_HTRANS,
      FIC_0_AHB_S_HWDATA    => FIC_0_AHB_S_HWDATA,
      FIC_0_AHB_S_HWRITE    => FIC_0_AHB_S_HWRITE,
      FIC_0_AHB_S_HRDATA    => FIC_0_AHB_S_HRDATA,
      FIC_0_AHB_S_HREADYOUT => FIC_0_AHB_S_HREADYOUT,
      FIC_0_AHB_S_HRESP     => FIC_0_AHB_S_HRESP,
      M3_RESET_N            => M3_RESET_N,
      MMUART_0_RXD          => MMUART_0_RXD,
      XTL                   => XTL,
      LOCK                  => LOCK,
      MMUART_0_TXD          => MMUART_0_TXD,
      MSS_RESET_N_M2F       => MSS_RESET_N_M2F,
      DEVRST_N              => devrst_n,
      POWER_ON_RESET_N      => POWER_ON_RESET_N,
      clk_100               => clk_100);

  tx_o <= MMUART_0_TXD;
  MMUART_0_RXD <= rx_i;

  M3_RESET_N <= '0';

  rst_n <= POWER_ON_RESET_N and LOCK and MSS_RESET_N_M2F;
  
  proc_led: process (clk_100)
    variable led1_cnt : natural;
  begin
    if rising_edge(clk_100) then
      if rst_n = '0' then
        led1_o <= '0';
        led2_o <= '0';
        led1_cnt := 0;
      else
        --  led1 is the led on the right (close to USB)
        if led1_cnt /= 0 then
          led1_cnt := led1_cnt - 1;
        else
          led1_o <= '0';
        end if;
        if cpu_err = '1' then
          led1_o <= '1';
          led1_cnt := 10_000_000;
        end if;

        --  led1_o <= counter (25);
        --  led2 is the led on the left (close to the hole).
        --  A led is on (emits light) when its input is 1.
        if state = DONE and counter (26) = '1' then
          led2_o <= '1';
        end if;
      end if;
    end if;
  end process;

  iram_addr <= iahb_addr (IRAM_LOG_SIZE - 1 downto 2);
  iram_data <= ahb_rdata;
  --  Compute ECC except for the last 4 words.
  iram_ecc_force <= b"000_0000" when iahb_addr(IRAM_LOG_SIZE - 1 downto 4) /= (IRAM_LOG_SIZE - 1 downto 4 => '1')
    else b"000_0001" when iahb_addr(2) = '0'
    else b"000_0011";
  iram_ecc <= f_calc_ecc (ahb_rdata) xor iram_ecc_force;

  proc_init: process (clk_100)
    constant delay : unsigned(27 downto 0) := x"0c0_0000"
    -- pragma translate off
    xor x"0c0_0010"
    -- pragma translate_on
    ;
  begin
    if rising_edge(clk_100) then
      iahb_start <= '0';
      cpu_rst_n <= '0';
      iram_we <= '0';

      if rst_n = '0' then
        state <= INIT_DELAY;
        counter <= (others => '0');
      else
        case state is
          when INIT_DELAY =>
            iahb_addr <= (others => '0');
            if counter = delay then
              state <= READ_NVM;
              counter <= (others => '0');
            else
              counter <= counter + 1;
            end if;
          when READ_NVM =>
            iahb_addr <= std_logic_vector (x"6000" & counter (13 downto 0) & "00");
            iahb_start <= '1';
            state <= WRITE_MEM;
          when WRITE_MEM =>
            if ahb_done = '1' then
              iram_we <= '1';
              if counter (IRAM_LOG_SIZE - 3 downto 0)= (IRAM_LOG_SIZE - 3 downto 0 => '1') then
                state <= DONE;
              else
                counter <= counter + 1;
                state <= READ_NVM;
              end if;
            end if;
          when DONE =>
            counter <= counter + 1;
            cpu_rst_n <= '1';
        end case;
      end if;
    end if;
  end process;

  ahb_start <= iahb_start when cpu_rst_n = '0' else (dwb_out.stb and dwb_out.cyc);
  ahb_addr  <= iahb_addr  when cpu_rst_n = '0' else dwb_out.adr;
  ahb_rw    <= '1'        when cpu_rst_n = '0' else not dwb_out.we;
  ahb_wdata <= dwb_out.dat;
  dwb_in.err <= '0';
  dwb_in.rty <= '0';
  dwb_in.stall <= '0'; --  FIXME
  dwb_in.dat <= ahb_rdata;
  dwb_in.ack <= cpu_rst_n and ahb_done;

  proc_ahb: process (clk_100)
  begin
    if rising_edge(clk_100) then
      FIC_0_AHB_S_HTRANS <= "00";
      FIC_0_AHB_S_HSEL <= '0';
      ahb_done <= '0';

      if rst_n = '0' then
        ahb_state <= STATE_AHB_IDLE;
      else
        case ahb_state is
          when STATE_AHB_IDLE =>
            if ahb_start = '1' then
              FIC_0_AHB_S_HADDR  <= ahb_addr;
              FIC_0_AHB_S_HTRANS <= "10";
              FIC_0_AHB_S_HWDATA <= ahb_wdata;
              FIC_0_AHB_S_HWRITE <= not ahb_rw;
              FIC_0_AHB_S_HSEL <= '1';
              ahb_state <= STATE_AHB_ADDR;
            end if;
          when STATE_AHB_ADDR =>
            --  Ready is expected to be 1.
            ahb_state <= STATE_AHB_DATA;
          when STATE_AHB_DATA =>
            if FIC_0_AHB_S_HREADYOUT = '1' then
              ahb_state <= STATE_AHB_IDLE;
              ahb_rdata <= FIC_0_AHB_S_HRDATA;
              ahb_done <= '1';
            end if;
        end case;
      end if; 
    end if;
  end process;

  FIC_0_AHB_S_HMASTLOCK <= '0';
  FIC_0_AHB_S_HSIZE <= "10"; --  Transfer size is always 32 bits.
  FIC_0_AHB_S_HREADY <= FIC_0_AHB_S_HREADYOUT;
end behav;
