#include <stdint.h>
#include "hydra_supervisor_regs.h"

extern void clear_bss(void);

#define SYS_RESET 0x40038048
#define UART_BASE 0x40000000
#define UART_RX  (UART_BASE | 0x00)
#define UART_TX  (UART_BASE | 0x00)
#define UART_DLR (UART_BASE | 0x00)
#define UART_DMR (UART_BASE | 0x04)
#define UART_LCR (UART_BASE | 0x0c)
#define UART_LSR (UART_BASE | 0x14)

#define SUPERVISOR_BASE 0x80000
#define SUPERVISOR ((volatile struct hydra_supervisor_regs *)SUPERVISOR_BASE)
#define WD_KEY 0xc0423bc9

static enum t_test
  {
   TEST_DRAM_DE,
   TEST_DRAM_SE,
   TEST_DRAM_SCRUB,
   TEST_IRAM_DE,
   TEST_IRAM_SE,
   TEST_IRAM_SCRUB,
   TEST_WD,
   TEST_CPU1_RECOVER,
   TEST_CPU2_RECOVER,
   TEST_CPU3_RECOVER,
   TEST_CPU1_FATAL,
   TEST_RAM,
   TEST_DONE,
} next_test;

static void
uart_init (void)
{
  /* Unreset UART 0 */
  *(volatile unsigned *)SYS_RESET &= 0xffffff7f;

  /* Select baud.  */
  *(volatile unsigned *)UART_LCR = 0x80;

  /* Set baudrate */
  *(volatile unsigned *)UART_DLR = 0x36;
  *(volatile unsigned *)UART_DMR = 0x00;

  /* start operation.  */
  *(volatile unsigned *)UART_LCR = 0x03;
}

static void
uart_raw_putc (unsigned char c)
{
  /* Wait until TEMT (transmit empty) is set.  */
  /* NB: bit 5 for TX not empty */
  do {
    SUPERVISOR->wd_key = WD_KEY;
  } while (!(*(volatile unsigned*)UART_LSR & 0x20));

  *(volatile unsigned *)UART_TX = c;
}

static int
uart_raw_getc (void)
{
  /* Check if DR (data ready, bit 0) is set.  */
  if (*(volatile unsigned*)UART_LSR & 0x1)
    return *(volatile unsigned *)UART_RX;
  else
    return -1;
}

static void
uart_putc (char c)
{
  if (c == '\n')
    uart_raw_putc ('\r');
  uart_raw_putc (c);
}

/* Put ONE hexadecimal digit.  */
static void
uart_put_hex_digit (unsigned val)
{
  uart_putc (val < 10 ? '0' + val : 'a' + val - 10);
}

static void
uart_put_hex8 (unsigned val)
{
  unsigned int i;
  for (i = 0; i < 8; i++) {
    uart_put_hex_digit(val >> 28);
    val <<= 4;
  }
}

static void
uart_puts (const char *s)
{
  while (*s)
    uart_putc (*s++);
}

static volatile unsigned ecc[4];

#define PAD_LEN 8
static volatile unsigned pad[8];

int
ram_test(void)
{
  int i;
  volatile unsigned char *p;

  for (i = 0; i < PAD_LEN; i++) {
    if (pad[i] != 0)
      return -1;
    pad[i] = 0xffffffff;
  }
  for (i = 0; i < PAD_LEN; i++) {
    p = (volatile unsigned char *)&pad[i];
    p[i & 3] = i;
  }
  for (i = 0; i < PAD_LEN; i++) {
    p = (volatile unsigned char *)&pad[i];
    unsigned v = 0xffffffff;
    v &= ~(0xff << (8 * (i & 3)));
    v |= i << (8 * (i & 3));
    if (pad[i] != v)
      return -2;
  }
  return 0;
}

/* ID must be 0-2 */
static void
force_divergence (unsigned id)
{
  unsigned v;

  /* Force divergence by executing different code.  */
  v = SUPERVISOR->force_divergence;
  if (v != (1 << id))
    uart_putc('D');

  /* Must be in lock-step.  */
  v = SUPERVISOR->cpu;
  uart_put_hex_digit(v & 0x0f);
}

static void
unreachable(void)
{
  while (1)
    uart_putc('$');
}

int
main (void)
{
  unsigned v;
  int c;

  SUPERVISOR->wd_key = WD_KEY;
  uart_init ();

  uart_puts ("Rst: ");
  v = SUPERVISOR->reset_cause;
  uart_put_hex_digit (v & 0x0f);
  uart_puts (", tst: ");
  uart_put_hex_digit (next_test & 0x0f);
  uart_putc('\n');

#if 0
  uart_puts("Go?");
  while (uart_raw_getc() != 'y')
    ;
  uart_putc('\n');
#endif

  switch (next_test) {
  case TEST_DRAM_DE:
    /* Add double-errors  */
    SUPERVISOR->dram_ecc_mask = 0x03;
    ecc[0] = 12;
    ecc[2] = 23;
    /* Add single-errors */
    SUPERVISOR->dram_ecc_mask = 0x01;
    ecc[1] = 56;
    ecc[3] = 78;
    /* Do not add errors anymore! */
    SUPERVISOR->dram_ecc_mask = 0x00;
    next_test = TEST_DRAM_SE;
    /* Order is important! */
    asm volatile ("");
    /* Trigger a single error */
    v = ecc[0];
    unreachable();
    break;

  case TEST_DRAM_SE:
    if (SUPERVISOR->dram_ecc_se != 0)
      unreachable();
    /* Trigger a single error */
    v = ecc[1];
    if (SUPERVISOR->dram_ecc_se != 1)
      unreachable();
    next_test = TEST_IRAM_DE;
    /* Fallthrough */

  case TEST_IRAM_DE:
    next_test = TEST_DRAM_SCRUB;
    /* Generate a fatal error */
    v = *(volatile unsigned *)0xfffc;
    unreachable();
    break;

  case TEST_DRAM_SCRUB:
    uart_putc('*');
    if (SUPERVISOR->dram_scrub_se != 0
        || SUPERVISOR->dram_scrub_de_curr != 0
        || SUPERVISOR->dram_scrub_de_last != 0
        || SUPERVISOR->dram_scrub_cycle != 0)
      unreachable();
    uart_puts("Scrub DRAM");
    SUPERVISOR->scrub_cfg = HYDRA_SUPERVISOR_REGS_SCRUB_CFG_DRAM_EN;
    next_test = TEST_IRAM_SE;
    do {
      SUPERVISOR->wd_key = WD_KEY;
    } while (SUPERVISOR->dram_scrub_cycle == 0);
    /* Because the scrubber is always running, the counter is always
       incrementing.  So the original value is kept. */
    uart_putc('\n');
    if (SUPERVISOR->dram_scrub_se != 1
        || SUPERVISOR->dram_scrub_de_last != 2) {
      uart_puts ("se:");
      uart_put_hex8 (SUPERVISOR->dram_scrub_se);
      uart_puts (", de:");
      uart_put_hex8 (SUPERVISOR->dram_scrub_de_last);
      unreachable();
    }
    /* Remove the errors */
    ecc[0] = 0;
    ecc[2] = 0;
    /* Fallthrough */

  case TEST_IRAM_SE:
    /* Generate a single error */
    v = *(volatile unsigned *)0xfff8;
    if (SUPERVISOR->iram_ecc_se != 1)
      unreachable();
    next_test = TEST_IRAM_SCRUB;
    /* Fallthrough */

  case TEST_IRAM_SCRUB:
    if (SUPERVISOR->iram_scrub_se != 0
        || SUPERVISOR->iram_scrub_de_curr != 0
        || SUPERVISOR->iram_scrub_de_last != 0
        || SUPERVISOR->iram_scrub_cycle != 0)
      unreachable();
    uart_puts("Scrub IRAM");
    SUPERVISOR->scrub_cfg = HYDRA_SUPERVISOR_REGS_SCRUB_CFG_IRAM_EN;
    next_test = TEST_WD;
    do {
      SUPERVISOR->wd_key = WD_KEY;
    } while (SUPERVISOR->iram_scrub_cycle == 0);
    uart_putc('\n');
    if (SUPERVISOR->iram_scrub_se != 1
        || SUPERVISOR->iram_scrub_de_last !=  2) {
      uart_puts ("se:");
      uart_put_hex8 (SUPERVISOR->iram_scrub_se);
      uart_puts (", de:");
      uart_put_hex8 (SUPERVISOR->iram_scrub_de_last);
      unreachable();
    }
    /* Fallthrough */

  case TEST_WD:
    /* Test watchdog - infinite loop */
    next_test = TEST_CPU1_RECOVER;
    while (1)
      v = pad[0];
    break;

  case TEST_CPU1_RECOVER:
  case TEST_CPU2_RECOVER:
  case TEST_CPU3_RECOVER:
    SUPERVISOR->wd_key = WD_KEY;
    force_divergence(next_test - TEST_CPU1_RECOVER);

    next_test++;

    /* Generate reset for recovery */
    SUPERVISOR->cpu = HYDRA_SUPERVISOR_REGS_CPU_RECOVERY;

    unreachable();
    break;

  case TEST_CPU1_FATAL:
    SUPERVISOR->wd_key = WD_KEY;
    /* Clear recovery */
    SUPERVISOR->cpu = 0;

    force_divergence(0);

    next_test = TEST_RAM;

    /* Force a data error in lock-step.
       This will generate a reset.  */
    v = SUPERVISOR->force_divergence;
    uart_put_hex_digit(v & 0x0f);

    unreachable();
    break;

  case TEST_RAM:
    SUPERVISOR->wd_key = WD_KEY;
    uart_puts ("Ram\n");
    v = ram_test();
    if (v != 0) {
      uart_put_hex_digit(v & 0x0f);
      unreachable();
    }

    next_test = TEST_DONE;
    /* Fallthrough */

  case TEST_DONE:
  default:
    next_test++;
    SUPERVISOR->wd_key = WD_KEY;
    SUPERVISOR->wd_period = 100000000;
    break;
  }

  while (1) {
    uart_puts ("i-nfo, r-ecover, l-oop\n");
    while ((c = uart_raw_getc()) == -1)
      ;
    switch (c) {
    case 'i':
      v = SUPERVISOR->wd_count;
      uart_puts ("wd count: ");
      uart_put_hex8 (v);
      uart_putc ('\n');
      uart_puts ("ecc dram: se:");
      uart_put_hex8 (SUPERVISOR->dram_scrub_se);
      uart_puts (", de:");
      uart_put_hex8 (SUPERVISOR->dram_scrub_de_last);
      uart_putc ('\n');
      uart_puts ("ecc iram: se:");
      uart_put_hex8 (SUPERVISOR->iram_scrub_se);
      uart_puts (", de:");
      uart_put_hex8 (SUPERVISOR->iram_scrub_de_last);
      uart_putc ('\n');
      break;
    case 'r':
      /* Generate reset for recovery */
      SUPERVISOR->cpu = HYDRA_SUPERVISOR_REGS_CPU_RECOVERY;

      unreachable();
    case 'l':
      while (1)
        ;
    default:
      break;
    }
  }
}
